{% extends "layout.html" %}
{% block box %}
  <h1>Build Failure :'(</h1>
  <p>
    Building of the artifact “{{ artifact }}” failed.  This is what we
    know about the problem:
  <pre class="traceback">{{ traceback }}</pre>
{% endblock %}
